#include "stdafx.h"

// iTunesCOMInterface.h and iTunesCOMInterface_i.c are part of
// the iTunes COM SDK which should be available from Apple:
//
// http://developer.apple.com/sdk/itunescomsdk.html
//
#include "..\..\Source\Third_Party\iTunesCOMWindowsSDK\iTunesCOMInterface.h"

namespace NLeoHelpers
{
	const TCHAR *SkipSortPrefixes(const TCHAR *szInput)
	{
		TCHAR *prefixes[] =
		{
			_T("the "),
			_T("a "),
			NULL
		};

		size_t prefLen;

		for (TCHAR **pszPrefix = prefixes; *pszPrefix != NULL; ++pszPrefix)
		{
			prefLen = _tcslen(*pszPrefix);

			if (0 == _tcsnicmp(*pszPrefix, szInput, prefLen))
			{
				const TCHAR *szOutput = szInput + prefLen;

				if (szOutput[0] != _T('\0'))
				{
					return szOutput;
				}
			}
		}

		return szInput;
	}

	void ToUpper(const std::basic_string< TCHAR > &strInput, std::basic_string< TCHAR > *pstrOutput)
	{
		if (NULL != pstrOutput)
		{
			pstrOutput->clear();

			std::transform(strInput.begin(), strInput.end(), std::back_inserter< std::basic_string< TCHAR > >(*pstrOutput), _totupper);
		}
	}

	void ToLower(const std::basic_string< TCHAR > &strInput, std::basic_string< TCHAR > *pstrOutput)
	{
		if (NULL != pstrOutput)
		{
			pstrOutput->clear();

			std::transform(strInput.begin(), strInput.end(), std::back_inserter< std::basic_string< TCHAR > >(*pstrOutput), _totlower);
		}
	}

	void ToUpper(std::basic_string< TCHAR > *pstr)
	{
		std::basic_string<TCHAR> temp = *pstr;
		NLeoHelpers::ToUpper(temp, pstr);
	}

	void ToLower(std::basic_string< TCHAR > *pstr)
	{
		std::basic_string<TCHAR> temp = *pstr;
		NLeoHelpers::ToLower(temp, pstr);
	}

	// delete[] the result.
	// Returns NULL on failure.
	TCHAR *VStringAllocAndFormat(const TCHAR *format, va_list args)
	{
		size_t bufferSize = 64;

		TCHAR *szResult = new(std::nothrow) TCHAR[bufferSize];

		while(true)
		{
			if (NULL == szResult)
			{
				::SetLastError(ERROR_NOT_ENOUGH_MEMORY);
				break;
			}

			int formatRes = _vsntprintf_s(szResult, bufferSize, _TRUNCATE, format, args);

			// formatRes negative on error; (bufferSize == formatRes) should be redundant for _vsntprintf.

			if ((0 > formatRes) || (bufferSize == formatRes))
			{
				delete [] szResult;
				bufferSize *= 2;
				szResult = new(std::nothrow) TCHAR[bufferSize];
			}
			else
			{
				break;
			}
		}

		return(szResult);
	}

	// delete[] the result.
	// Returns NULL on failure.
	TCHAR *StringAllocAndFormat(const TCHAR *format, ...)
	{
		va_list args;
		va_start(args, format);
		TCHAR *result = VStringAllocAndFormat(format, args);
		va_end(args);

		return(result);
	}
}

IITFileOrCDTrack *GetITFileOrCDTrackFromVariant(VARIANT &varTrack)
{
	IUnknown *pUnkTrack = NULL;

	switch(varTrack.vt)
	{
	case(VT_DISPATCH):
		pUnkTrack = varTrack.pdispVal;
		break;
	case(VT_DISPATCH | VT_BYREF):
		if (varTrack.ppdispVal != NULL)
		{
			pUnkTrack = *varTrack.ppdispVal;
		}
		break;
	case(VT_UNKNOWN):
		pUnkTrack = varTrack.punkVal;
		break;
	case(VT_UNKNOWN | VT_BYREF):
		if (varTrack.ppunkVal != NULL)
		{
			pUnkTrack = *varTrack.ppunkVal;
		}
		break;
	default:
		break;
	}

	IITFileOrCDTrack *pTrack = NULL;

	if (pUnkTrack == NULL
	||	S_OK != pUnkTrack->QueryInterface(IID_IITFileOrCDTrack, reinterpret_cast<void **>(&pTrack)))
	{
		pTrack = NULL;
	}

	return pTrack;
}

class CTrackInfo
{
private:
	_bstr_t m_bstrArtist;
	_bstr_t m_bstrAlbumArtist;
	_bstr_t m_bstrAlbum;
	_bstr_t m_bstrSortAlbum;
	_bstr_t m_bstrName;
	long m_lYear;
	bool m_bOkay;
	IITFileOrCDTrack *m_pTrack;

	CTrackInfo &operator=(const CTrackInfo &rhs);
	CTrackInfo(const CTrackInfo &rhs);

public:
	CTrackInfo(VARIANT &varTrack)
	: m_lYear(0)
	, m_bOkay(false)
	, m_pTrack(NULL)
	{
		m_pTrack = GetITFileOrCDTrackFromVariant(varTrack);

		if (m_pTrack != NULL)
		{
			if (S_OK == m_pTrack->get_Artist(m_bstrArtist.GetAddress())
			&&	S_OK == m_pTrack->get_AlbumArtist(m_bstrAlbumArtist.GetAddress())
			&&	S_OK == m_pTrack->get_Album(m_bstrAlbum.GetAddress())
			&&	S_OK == m_pTrack->get_SortAlbum(m_bstrSortAlbum.GetAddress())
			&&	S_OK == m_pTrack->get_Name(m_bstrName.GetAddress())
			&&	S_OK == m_pTrack->get_Year(&m_lYear)
			&&	m_bstrArtist.length() != 0
			&&	m_bstrAlbum.length() != 0
			&&	m_bstrName.length() != 0
			&&	m_lYear > 999 && m_lYear < 10000)
			{
				m_bOkay = true;
			}
		}
	}

	virtual ~CTrackInfo()
	{
		if (m_pTrack != NULL)
		{
			m_pTrack->Release();
			m_pTrack = NULL;
		}
	}

	bool IsOkay() const { return m_bOkay; }

	const wchar_t *GetArtist() const { return m_bOkay ? static_cast<const wchar_t *>(m_bstrArtist) : NULL; }
	const wchar_t *GetAlbum() const  { return m_bOkay ? static_cast<const wchar_t *>(m_bstrAlbum)  : NULL; }
	const wchar_t *GetName() const   { return m_bOkay ? static_cast<const wchar_t *>(m_bstrName)   : NULL; }

	bool HasAlbumArtist() const { return (m_bOkay && m_bstrAlbumArtist.length() != 0); }
	bool HasSortAlbum() const	{ return (m_bOkay && m_bstrSortAlbum.length()   != 0); }

	const wchar_t *GetOptionalAlbumArtist() const { return HasAlbumArtist() ? static_cast<const wchar_t *>(m_bstrAlbumArtist) : NULL; }
	const wchar_t *GetOptionalSortAlbum() const	  { return HasSortAlbum()   ? static_cast<const wchar_t *>(m_bstrSortAlbum)   : NULL; }

	const wchar_t *GetAlbumArtistOrArtist() const { return HasAlbumArtist() ? GetOptionalAlbumArtist() : GetArtist(); }

	long GetYear() const					{ return m_bOkay ? m_lYear : 0; }

	bool SetSortAlbum(const wchar_t *szNewSortAlbum, bool bPretend)
	{
		// Result should indicates that a tag was changed (or would have been if not pretending).
		// If Result is false then either the tag was already as desired or there was an error.
		bool bResult = false;

		if (m_bOkay
		&&	szNewSortAlbum != NULL)
		{
			std::wstring strCurrentSortAlbum = HasSortAlbum() ? GetOptionalSortAlbum() : L"";

			if (0 != wcscmp(szNewSortAlbum, strCurrentSortAlbum.c_str()))
			{
				m_bstrSortAlbum = szNewSortAlbum;

				if (bPretend)
				{
					wprintf(L"Would change \"%s\" SortAlbum \"%s\" to \"%s\"\n", GetAlbum(), strCurrentSortAlbum.c_str(), szNewSortAlbum);
					bResult = true;
				}
				else if (S_OK == m_pTrack->put_SortAlbum(m_bstrSortAlbum))
				{
					_bstr_t bstrReadBackSortAlbum;
					
					if (S_OK == m_pTrack->get_SortAlbum(bstrReadBackSortAlbum.GetAddress()))
					{
						// iTunes will remove prefixes by itself so report what it really set.

						if (bstrReadBackSortAlbum.length() == 0)
						{
							szNewSortAlbum = L"";
						}
						else
						{
							szNewSortAlbum = bstrReadBackSortAlbum;
						}
					}

					wprintf(L"Changed \"%s\" SortAlbum \"%s\" to \"%s\"\n", GetAlbum(), strCurrentSortAlbum.c_str(), szNewSortAlbum);
					bResult = true;
				}
				else
				{
					wprintf(L"Failed to change \"%s\" SortAlbum \"%s\" to \"%s\"\n", GetAlbum(), strCurrentSortAlbum.c_str(), szNewSortAlbum);
				}
			}
		}

		return bResult;
	}

//	IITFileOrCDTrack *GetObject() { return m_bOkay ? m_pTrack : NULL; }
};

class CAlbumsInfo
{
private:
	std::map< std::wstring, std::map< std::wstring, long > > m_artistMap;

	CAlbumsInfo &operator=(const CAlbumsInfo &rhs);
	CAlbumsInfo(const CAlbumsInfo &rhs);

public:
	CAlbumsInfo()
	{
	}

	virtual ~CAlbumsInfo()
	{
	}

	void Update(const CTrackInfo &trackInfo)
	{
		if (trackInfo.IsOkay())
		{
			std::wstring strArtist = trackInfo.GetAlbumArtistOrArtist();
			std::wstring strArtistLower;

			NLeoHelpers::ToLower(strArtist, &strArtistLower);

			std::wstring strAlbum = trackInfo.GetAlbum();
			std::wstring strAlbumLower;
			NLeoHelpers::ToLower(strAlbum, &strAlbumLower);

			long lTrackYear = trackInfo.GetYear();

			if (m_artistMap[strArtistLower][strAlbumLower] < lTrackYear)
			{
				m_artistMap[strArtistLower][strAlbumLower] = lTrackYear;

				//wprintf(L"%s - %s --> %ld\n", strArtist.c_str(), strAlbum.c_str(), lTrackYear);
				wprintf(L".");
			}
		}
	}

	long GetAlbumYear(const CTrackInfo &trackInfo)
	{
		long lAlbumYear = 0;

		if (trackInfo.IsOkay())
		{
			std::wstring strArtist = trackInfo.GetAlbumArtistOrArtist();
			std::wstring strArtistLower;

			NLeoHelpers::ToLower(strArtist, &strArtistLower);

			std::wstring strAlbum = trackInfo.GetAlbum();
			std::wstring strAlbumLower;
			NLeoHelpers::ToLower(strAlbum, &strAlbumLower);

			long lTrackYear = trackInfo.GetYear();

			lAlbumYear = m_artistMap[strArtistLower][strAlbumLower];

			if (lTrackYear > lAlbumYear)
			{
				lAlbumYear = lTrackYear;

				m_artistMap[strArtistLower][strAlbumLower] = lTrackYear;

				wprintf(L"WARNING: Album Year has changed! %s - %s --> %ld\n", strArtist.c_str(), strAlbum.c_str(), lTrackYear);
			}
		}

		return lAlbumYear;
	}
};

int wmain(int argc, wchar_t* argv[])
{
	wprintf(L"\niPodYearSort (C)2007 Leo Davidson\nhttp://www.PretentiousName.com\n\n");

	bool bHelp = false;
	bool bDescribe = false;
	bool bPretend = false;
	bool bWrite = false;
	bool bClear = false;

	if (argc != 2)
	{
		bHelp = true;
	}
	else if (0 == _wcsicmp(L"/Describe", argv[1]))
	{
		bDescribe = true;
	}
	else if (0 == _wcsicmp(L"/PretendWrite", argv[1]))
	{
		bPretend = true;
		bWrite = true;
	}
	else if (0 == _wcsicmp(L"/RealWrite", argv[1]))
	{
		bWrite = true;
	}
	else if (0 == _wcsicmp(L"/PretendClear", argv[1]))
	{
		bPretend = true;
		bClear = true;
	}
	else if (0 == _wcsicmp(L"/RealClear", argv[1]))
	{
		bClear = true;
	}
	else
	{
		bHelp = true;
	}

	if (bHelp)
	{
		wprintf(L"This program must be run from a Command Prompt.\n\n");
		wprintf(L"Run with exaclty one of the following arguments:\n\n");
		wprintf(L"    /Describe: Show a detailed description of what this program does.\n");
		wprintf(L"/PretendWrite: Tells you what would happen if new Sort tags were written.\n");
		wprintf(L"   /RealWrite: Really writes new Sort tags.\n");
		wprintf(L"/PretendClear: Tells you what would happen if the Sort tags were cleared.\n");
		wprintf(L"   /RealClear: Really clears the Sort tags.\n");

		Sleep(5000); // In case people double-click it. Gives them time to read the text.
	}
	else if (bDescribe)
	{
		wprintf(L"This program instructs iTunes to adjust the SortAlbum tags on your music\n");
		wprintf(L"library so that albums are sorted by year when viewed on an iPod.\n");
		wprintf(L"\n");
		wprintf(L"In addition, compilation albums will be sorted below non-compilation albums\n");
		wprintf(L"so that when you view an Artist on your iPod you will see all of their main\n");
		wprintf(L"albums followed by the compilations they appear on, rather than a mixture\n");
		wprintf(L"of the two.\n");
		wprintf(L"\n");
		wprintf(L"Your music files must have Year tags for this program to work. Any files\n");
		wprintf(L"without Year tags will be skipped.\n");
		wprintf(L"\n");
		wprintf(L"An album is considered a compilation if its AlbumArtist tag starts\n");
		wprintf(L"with \"Various\". Compilations are *not* detected by the \"Part of a\n");
		wprintf(L"Compilation Album\" tag in iTunes because this tag is not available\n");
		wprintf(L"through the iTunes COM API. :-(\n");
		wprintf(L"\n");
		wprintf(L"When you add new music to your iTunes library you should re-run this tool\n");
		wprintf(L"to adjust the sort tags on the new files.\n");
		wprintf(L"\n");
		wprintf(L"Be prepared for a long iPod sync after using this tool for the first time\n");
		wprintf(L"as iTunes will re-sync the complete files even though only the tags have\n");
		wprintf(L"changed. Don't worry, the next time you run the tool it will only adjust\n");
		wprintf(L"the files which need it, leaving alone those it has already done.\n");
		wprintf(L"\n");
		wprintf(L"This program is free and open-source. See http://www.PretentiousName.com\n");
		wprintf(L"\n");
	}
	else if (S_OK == ::CoInitializeEx(NULL, COINIT_MULTITHREADED))
	{
		IiTunes *pITunes = NULL;
		IITLibraryPlaylist *pLibraryPlaylist = NULL;
		IITTrackCollection *pTrackCollection = NULL;
		IUnknown *pUnkTrackEnum = NULL;
		IEnumVARIANT *pTrackEnum = NULL;
		VARIANT_BOOL bVersionCheck = VARIANT_FALSE;
    
		wprintf(L"\nInterfacing with iTunes...\n\n");

	    if (S_OK != ::CoCreateInstance(CLSID_iTunesApp, NULL, CLSCTX_LOCAL_SERVER, IID_IiTunes, reinterpret_cast<void **>(&pITunes))
		||	S_OK != pITunes->CheckVersion(1,1,&bVersionCheck)
		||	!bVersionCheck
		||	S_OK != pITunes->get_LibraryPlaylist(&pLibraryPlaylist)
		||	S_OK != pLibraryPlaylist->get_Tracks(&pTrackCollection)
		||	S_OK != pTrackCollection->get__NewEnum(&pUnkTrackEnum)
		||	S_OK != pUnkTrackEnum->QueryInterface(IID_IEnumVARIANT, reinterpret_cast<void **>(&pTrackEnum)))
		{
			wprintf(L"Could not get track list. (Incompatible version of iTunes?)\n");
		}
		else
		{
			int iInvalidCount = 0;

			CAlbumsInfo albumsInfo;

			VARIANT varTrack;
			VariantInit(&varTrack);

			if (!bClear)
			{
				wprintf(L"\nCalculating album years...\n\n");

				while (S_OK == pTrackEnum->Next(1, &varTrack, NULL))
				{
					CTrackInfo trackInfo(varTrack);

					if (trackInfo.IsOkay())
					{
						albumsInfo.Update(trackInfo);
					}

					VariantClear(&varTrack);
				}

				VariantClear(&varTrack);
			}

			if (bPretend)
			{
				wprintf(L"\n\nPRETENDING to update SortAlbum fields...\n\n");
			}
			else
			{
				wprintf(L"\n\nUpdating SortAlbum fields...\n\n");
			}

			if (S_OK != pTrackEnum->Reset())
			{
				wprintf(L"Error resetting enumerator.\n");
			}
			else
			{
				bool bAtLeastOneValid = false;
				bool bAtLeastOneChanged = false;

				while (S_OK == pTrackEnum->Next(1, &varTrack, NULL))
				{
					CTrackInfo trackInfo(varTrack);

					if (!trackInfo.IsOkay())
					{
						++iInvalidCount;
					}
					else
					{
						bAtLeastOneValid = true;

						std::wstring strAlbumNoPrefix = NLeoHelpers::SkipSortPrefixes(trackInfo.GetAlbum());

						std::wstring strNewSortAlbum;
						bool bSuccess = true;

						if (bClear)
						{
						//	if (0 != wcscmp(strAlbumNoPrefix.c_str(), trackInfo.GetAlbum()))
						//	{
						//		strNewSortAlbum = strAlbumNoPrefix; // Try to set things back to how iTunes would've done them.
						//	}
							// Else leave it blank as we want to clear the AlbumSort field.
						}
						else
						{
							const wchar_t *szVAPrefix = L"";

							if (trackInfo.HasAlbumArtist()
							&&	0 == _wcsnicmp(trackInfo.GetOptionalAlbumArtist(), L"Various", 7))
							{
								szVAPrefix = L"9999 ";
							}

							long lYear = albumsInfo.GetAlbumYear(trackInfo);

							wchar_t *szNewSortAlbum = NLeoHelpers::StringAllocAndFormat(L"%s%ld %s",
								szVAPrefix, lYear, strAlbumNoPrefix.c_str());

							if (szNewSortAlbum == NULL)
							{
								bSuccess = false;
							}
							else
							{
								strNewSortAlbum = szNewSortAlbum;
								delete [] szNewSortAlbum;
							}
						}

						if (bSuccess)
						{
							if (trackInfo.SetSortAlbum(strNewSortAlbum.c_str(), bPretend))
							{
								bAtLeastOneChanged = true;
							}
						}
					}

					VariantClear(&varTrack);
				}

				if (!bAtLeastOneValid)
				{
					wprintf(L"No suitable tracks were found. Do your music files have Year tags?\n");
				}
				else if (!bAtLeastOneChanged)
				{
					wprintf(L"No changes were performed.\n");
				}
				else
				{
					wprintf(L"\nDone.\n");
				}

				if (bAtLeastOneValid && iInvalidCount>0)
				{
					wprintf(L"\nWarning: %d of your tracks were ignored by the tool due to missing tags.\nYear tags, in particular, are required.\n", iInvalidCount);
				}
			}

			VariantClear(&varTrack);
		}

		if (pTrackEnum != NULL)			{ pTrackEnum->Release();		}
		if (pUnkTrackEnum != NULL)		{ pUnkTrackEnum->Release();		}
		if (pTrackCollection != NULL)	{ pTrackCollection->Release();	}
		if (pLibraryPlaylist != NULL)	{ pLibraryPlaylist->Release();	}
		if (pITunes != NULL)			{ pITunes->Release();			}

		::CoUninitialize();
	}

	return 0;
}
